#include<iostream.h>
#include<string>


#include "ipaddress.h"
#include "router.h"
#include "packet.h"
#include "flow.h"
#include <vector>
#include <sys/time.h>

DataLink **dataLinks;
Router **routers;
int routerC;
int dataLinkC;

#define MAX_ROUTERS 15
#define MAX_DATALINKS 15

vector<flow> flows;

DataLink *findDataLink(IPAddress a)
{
  for(int i = 0; i < dataLinkC; i++)
    {
      if(dataLinks[i]->net() == a)
	return dataLinks[i];
    }
  return NULL;
}

Router *findRouter(string name)
{
  for(int i = 0; i < routerC; i++)
    {
      if(routers[i]->name() == name)
	return routers[i];
    }
  return NULL;
}
void parse_router()
{
  string name;
  cin >> name;
  // cout << "working on router " << name << "...";
  string token;
  cin  >> token;
  //cout << " Router " << name << endl;
  
  Router * theRouter = new Router(name);
  while(token != "end-router")
    {
      //      cout << " tk[" << token << "] ";
      if(token == "interface")
	{
	  string network_address;
	  string IP_address;
	  cin >> network_address >> IP_address;
	  DataLink * DL = findDataLink(network_address);
	  if(DL)
	    theRouter->addInterface(DL, IPAddress(IP_address));
	  else
	    cout << " ERROR: Data link " << network_address << " not defined." << endl;
	}
      else if (token == "route")
	{
	  //	  cout << " **parseRoute** ";
	  string destination_network;
	  string next_hop;
	  int metric;
	  cin >> destination_network;
	  cin >> next_hop;
	  cin >> metric;
	  theRouter->addRoute(IPAddress(destination_network), IPAddress(next_hop), metric);
	}
      else
	cout << "    Parse error on token " << token << endl;
      cin >> token; 
      //cout << " TK[" << token << "] ";
    }
  routers[routerC] = theRouter;
  routerC++;
  //cout << " added router " << name << endl;
}

void parse_datalink()
{
  string network_address;
  int bandwidth;
  int delay;
  cin >> network_address;
  cin >> bandwidth;
  cin >> delay;

  dataLinks[dataLinkC] = new DataLink(IPAddress(network_address), bandwidth, delay);
  dataLinkC++;
}

void parse_flow()
{
  string srcRouter, src, dst;
  int pktsize, pktrate, starttime, duration;


  cin >> srcRouter >> src >> dst >> pktsize >> pktrate >> starttime >> duration;
  
  flow f(findRouter(srcRouter), IPAddress(src), IPAddress(dst), pktsize, pktrate, starttime, duration);
  //cout << " flow rate is " << pktrate << endl;
  flows.push_back(f);
}

void parse_runsim()
{
  int sliceWidth, sliceCount;
  cin >> sliceWidth >> sliceCount;  
  cout << " Running simulation: " << sliceCount << " slices of " << sliceWidth << "ms each." << endl;

  /* initialize the random number generator */
  struct timeval tv;
  struct timezone tz;
  gettimeofday(&tv, &tz);
  srand(tv.tv_usec % 100);


  for(int i =0; i < (sliceCount * sliceWidth); i+=sliceWidth)
    {
      cout << " ----------  " << i << " ms --------- " << endl;
      for(unsigned int q = 0; q < flows.size(); q++)
	{
	  if(flows[q].starttime() <= i && flows[q].duration() > 0)
	    {
	      flows[q].adjDuration(-sliceWidth);

	      int pktCount = int( flows[q].pktrate() * (sliceWidth * .001));
	      // cout << "Delivering " << pktCount << " packets to " << flows[q].srcRouter()->name() << " rate: " << flows[q].pktrate() << endl;
	      for(int l = 0; l < pktCount; l++)
		{
		  Packet *p = new Packet(flows[q].src(), flows[q].dst(), flows[q].srcRouter()->name() , flows[q].pktsize());
		  flows[q].packets.push_back(p);
		  flows[q].srcRouter()->deliverPacket(p);
		}
	    }
	}
      for(int j = 0; j < routerC; j++)
	{
	  routers[j]->rippleQueue();
	}
      for(int j = 0; j < routerC; j++)
	{
	  routers[j]->runTimeSlice(sliceWidth);
	}    

      for(int j =0; j < dataLinkC; j++)
	{
	  dataLinks[j]->runTimeSlice(sliceWidth);
	}
    }
  cout << "  ** Flow Analysis ** " << endl;
  for(unsigned int i = 0; i < flows.size(); i ++)
    {
      cout << "Flow from " << flows[i].srcRouter()->name() << " to " << flows[i].dst().toString() << ": " << endl;

       int pktCount = flows[i].packets.size();
       int deliveredCount = 0;
       int netDelay = 0;
       for(int j = 0; j < pktCount; j++)
	 {
	   if(flows[i].packets[j]->delivered())
	     {
	       netDelay += flows[i].packets[j]->delay();
	       deliveredCount++;
	     }
	 }

       if(pktCount != 0)
	 {
	   cout << "   Packets delivered: " << (deliveredCount * 100 / pktCount) << "%" << endl;
	   cout << "   Average Delay    : " << (netDelay / pktCount) << "ms" << endl;
	 }
       else
	 {
	   cout << "   Error: No packets were delivered." << endl;
	 }

    }
}

void parse_simulation()
{
  flows.clear();
  string token;
  cin >> token;
  while(token != "end-simulation")
    {
      if(token == "flow")
	{
	  parse_flow();
	}
      else if(token == "run")
	{
	  parse_runsim();
	}
      else
	{
	  cout << "Configreader: Parse error in simulation on token " << token << endl;
	}
      cin >> token;
    }

}

void parse_packet()
{
  string routerName, src, dst, payload;
  cin >> routerName >> src >> dst >> payload;
  Router *rtr = findRouter(routerName);
  
  if(rtr==0)
    {
      cout << " Error: no router [" << routerName << "] found." << endl;
      cout.flush();
    }
  else
    {
      cout.flush();
      Packet *p = new Packet(IPAddress(src), IPAddress(dst), payload);
      //p->setRecordRoute(true);
      p->appendPayload("\n");
      rtr->deliverPacket(p);
      //delete p;
      
    }
  //cin >> tToken;

}

/* "traffic" sections of network config files
   pass around individual packets on the network.
 
   Flows are better; use flows.

*/

void parse_traffic()
{
  string tToken;
  cin >> tToken;
  while(tToken != "end-traffic")
    {
      if(tToken == "packet")
	{
	  parse_packet();
	}
      else
	{
	  cout << "Parse error on token " << tToken << endl;
	}
      cin >> tToken;
    }

  /* perform the simulation */
  for(int i =0; i < 100; i+=10)
    {
      cout << " ----------  " << i << " vSecs --------- " << endl;
      
      for(int j = 0; j < routerC; j++)
	{
	  routers[j]->rippleQueue();
	}
      for(int j = 0; j < routerC; j++)
	{
	  routers[j]->runTimeSlice(10);
	}    
      for(int j =0; j < dataLinkC; j++)
	{
	  dataLinks[j]->runTimeSlice(10);
	}    
    }

}

int main()
{

  routers = new Router*[MAX_ROUTERS];
  dataLinks = new DataLink*[MAX_DATALINKS];
  routerC =0;
  dataLinkC=0;

  string token;
  cin >> token;
  while(token != "end-network")
    {
      if(token == "router")
	parse_router();
      else if (token == "datalink")
	parse_datalink();
      else
	cout << " ** Parse Error on Token [" << token << "]" << endl;
      cin >> token;
    }

  /*

    // print debugging info: datalinks, routers, routing table entries
    for(int i = 0; i < dataLinkC; i++)
    {
    dataLinks[i]->listHosts();
    }
    for(int i =0; i < routerC; i++)
    {
    routers[i]->printInterfaces();
    routers[i]->printRoutes();
    }
  */
  
  cin >> token;
  while(token != "end")
    {
      if(token == "traffic")
	{
	  parse_traffic();
	}
      else if (token == "simulation")
	{
	  parse_simulation();
	}
      else
	{
	  cout << "Configreader: parse error on token " << token << endl;
	}
      cin >> token;
    }
		
  return 0;
}















